@@ -52,26 +52,15 @@
 #endif
 #define G_LOG_DOMAIN "PpsJobs"
 
-enum {
-	PROP_0,
-	PROP_DOCUMENT,
-};
-
-enum {
-	CANCELLED,
-	FINISHED,
-	LAST_SIGNAL
-};
 
 enum {
 	FIND_UPDATED,
 	FIND_LAST_SIGNAL
 };
 
-static guint job_signals[LAST_SIGNAL] = { 0 };
+
 static guint job_find_signals[FIND_LAST_SIGNAL] = { 0 };
 
-G_DEFINE_ABSTRACT_TYPE (PpsJob, pps_job, G_TYPE_OBJECT)
 G_DEFINE_TYPE (PpsJobLinks, pps_job_links, PPS_TYPE_JOB)
 G_DEFINE_TYPE (PpsJobAttachments, pps_job_attachments, PPS_TYPE_JOB)
 G_DEFINE_TYPE (PpsJobAnnots, pps_job_annots, PPS_TYPE_JOB)
@@ -80,244 +69,11 @@ G_DEFINE_TYPE (PpsJobPageData, pps_job_page_data, PPS_TYPE_JOB)
 G_DEFINE_TYPE (PpsJobThumbnailTexture, pps_job_thumbnail_texture, PPS_TYPE_JOB)
 G_DEFINE_TYPE (PpsJobFonts, pps_job_fonts, PPS_TYPE_JOB)
 G_DEFINE_TYPE (PpsJobLoad, pps_job_load, PPS_TYPE_JOB)
-G_DEFINE_TYPE (PpsJobSave, pps_job_save, PPS_TYPE_JOB)
 G_DEFINE_TYPE (PpsJobFind, pps_job_find, PPS_TYPE_JOB)
 G_DEFINE_TYPE (PpsJobLayers, pps_job_layers, PPS_TYPE_JOB)
 G_DEFINE_TYPE (PpsJobExport, pps_job_export, PPS_TYPE_JOB)
 G_DEFINE_TYPE (PpsJobPrint, pps_job_print, PPS_TYPE_JOB)
 
-/* PpsJob */
-static void
-pps_job_init (PpsJob *job)
-{
-	job->cancellable = g_cancellable_new ();
-}
-
-static void
-pps_job_dispose (GObject *object)
-{
-	PpsJob *job = PPS_JOB (object);
-
-	g_debug ("disposing %s (%p)", PPS_GET_TYPE_NAME(job), job);
-
-	g_clear_object (&job->document);
-	g_clear_object (&job->cancellable);
-	g_clear_error (&job->error);
-
-	(* G_OBJECT_CLASS (pps_job_parent_class)->dispose) (object);
-}
-
-static void
-pps_job_set_property (GObject      *object,
-		     guint         prop_id,
-		     const GValue *value,
-		     GParamSpec   *pspec)
-{
-	PpsJob *job = PPS_JOB (object);
-
-	switch (prop_id) {
-	case PROP_DOCUMENT:
-		job->document = g_value_dup_object (value);
-		break;
-	default:
-		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-	}
-}
-
-static void
-pps_job_class_init (PpsJobClass *klass)
-{
-	GObjectClass *g_object_class = G_OBJECT_CLASS (klass);
-
-	g_object_class->dispose = pps_job_dispose;
-	g_object_class->set_property = pps_job_set_property;
-
-	job_signals[CANCELLED] =
-		g_signal_new ("cancelled",
-			      PPS_TYPE_JOB,
-			      G_SIGNAL_RUN_LAST,
-			      G_STRUCT_OFFSET (PpsJobClass, cancelled),
-			      NULL, NULL,
-			      g_cclosure_marshal_VOID__VOID,
-			      G_TYPE_NONE, 0);
-	job_signals [FINISHED] =
-		g_signal_new ("finished",
-			      PPS_TYPE_JOB,
-			      G_SIGNAL_RUN_FIRST,
-			      G_STRUCT_OFFSET (PpsJobClass, finished),
-			      NULL, NULL,
-			      g_cclosure_marshal_VOID__VOID,
-			      G_TYPE_NONE, 0);
-
-	g_object_class_install_property (g_object_class,
-					 PROP_DOCUMENT,
-					 g_param_spec_object ("document",
-							      "Document",
-							      "The document",
-							      PPS_TYPE_DOCUMENT,
-							      G_PARAM_WRITABLE |
-							      G_PARAM_CONSTRUCT_ONLY |
-							      G_PARAM_STATIC_STRINGS));
-}
-
-static gboolean
-emit_finished (PpsJob *job)
-{
-	g_debug ("emit finished for %s (%p)", PPS_GET_TYPE_NAME (job), job);
-
-	job->idle_finished_id = 0;
-
-	if (job->cancelled)
-		g_debug ("%s (%p) job was cancelled, do not emit finished", PPS_GET_TYPE_NAME (job), job);
-	else
-		g_signal_emit (job, job_signals[FINISHED], 0);
-
-	return G_SOURCE_REMOVE;
-}
-
-static void
-pps_job_emit_finished (PpsJob *job)
-{
-	if (g_cancellable_is_cancelled (job->cancellable)) {
-		g_debug ("%s (%p) job was cancelled, returning", PPS_GET_TYPE_NAME (job), job);
-		return;
-	}
-
-	job->finished = TRUE;
-
-	job->idle_finished_id =
-		g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
-				 (GSourceFunc)emit_finished,
-				 g_object_ref (job),
-				 (GDestroyNotify)g_object_unref);
-}
-
-gboolean
-pps_job_run (PpsJob *job)
-{
-	PpsJobClass *class = PPS_JOB_GET_CLASS (job);
-	gboolean ret;
-
-	PPS_PROFILER_START (PPS_GET_TYPE_NAME (job), g_strdup("running"));
-	ret = class->run (job);
-	PPS_PROFILER_STOP ();
-	return ret;
-}
-
-void
-pps_job_cancel (PpsJob *job)
-{
-	if (job->cancelled)
-		return;
-
-	g_debug ("job %s (%p) cancelled", PPS_GET_TYPE_NAME (job), job);
-
-	/* This should never be called from a thread */
-	job->cancelled = TRUE;
-	g_cancellable_cancel (job->cancellable);
-
-        if (job->finished && job->idle_finished_id == 0)
-                return;
-
-	g_signal_emit (job, job_signals[CANCELLED], 0);
-}
-
-void
-pps_job_failed (PpsJob       *job,
-	       GQuark       domain,
-	       gint         code,
-	       const gchar *format,
-	       ...)
-{
-	va_list args;
-	gchar  *message;
-
-	if (job->failed || job->finished)
-		return;
-
-	g_debug ("job %s (%p) failed", PPS_GET_TYPE_NAME (job), job);
-
-	job->failed = TRUE;
-
-	va_start (args, format);
-	message = g_strdup_vprintf (format, args);
-	va_end (args);
-
-	job->error = g_error_new_literal (domain, code, message);
-	g_free (message);
-
-	pps_job_emit_finished (job);
-}
-
-/**
- * pps_job_failed_from_error: (rename-to pps_job_failed)
- * @job: an #PpsJob
- * @error: a #GError
- */
-void
-pps_job_failed_from_error (PpsJob  *job,
-			  GError *error)
-{
-	if (job->failed || job->finished)
-		return;
-
-	g_debug ("job %s (%p) failed", PPS_GET_TYPE_NAME (job), job);
-
-	job->failed = TRUE;
-	job->error = g_error_copy (error);
-
-	pps_job_emit_finished (job);
-}
-
-void
-pps_job_succeeded (PpsJob *job)
-{
-	if (job->finished)
-		return;
-
-	g_debug ("job %s (%p) succeeded", PPS_GET_TYPE_NAME (job), job);
-
-	job->failed = FALSE;
-	pps_job_emit_finished (job);
-}
-
-gboolean
-pps_job_is_finished (PpsJob *job)
-{
-	return job->finished;
-}
-
-/**
- * pps_job_is_succeeded:
- * @job: a #PpsJob
- * @error (nullable): (transfer none): the error to set if the job failed
- *
- * Returns: whether the job succeed
- */
-gboolean
-pps_job_is_succeeded (PpsJob   *job, GError **error)
-{
-	g_return_val_if_fail (job != NULL, FALSE);
-
-	if (job->failed && error)
-		*error = job->error;
-
-	return !job->failed;
-}
-
-/**
- * pps_job_get_document:
- * @job: an #PpsJob
- *
- * Returns: (transfer none): The #PpsDocument of this job.
- */
-PpsDocument *
-pps_job_get_document (PpsJob	 *job)
-{
-	g_return_val_if_fail (PPS_IS_JOB (job), NULL);
-
-	return job->document;
-}
 
 /* PpsJobLinks */
 static void
@@ -355,7 +111,7 @@ fill_page_labels (GListModel *model, PpsJob *job)
 		if (!link)
 			continue;;
 
-		document_links = PPS_DOCUMENT_LINKS (job->document);
+		document_links = PPS_DOCUMENT_LINKS (pps_job_get_document (job));
 		page_label = pps_document_links_get_link_page_label (document_links, link);
 		if (!page_label)
 			continue;
@@ -379,9 +135,9 @@ pps_job_links_run (PpsJob *job)
 
 	g_debug ("running links job");
 
-	pps_document_doc_mutex_lock ();
-	job_links->model = pps_document_links_get_links_model (PPS_DOCUMENT_LINKS (job->document));
-	pps_document_doc_mutex_unlock ();
+	pps_document_doc_mutex_lock (pps_job_get_document (job));
+	job_links->model = pps_document_links_get_links_model (PPS_DOCUMENT_LINKS (pps_job_get_document (job)));
+	pps_document_doc_mutex_unlock (pps_job_get_document (job));
 
 	fill_page_labels (job_links->model, job);
 
@@ -453,10 +209,10 @@ pps_job_attachments_run (PpsJob *job)
 
 	g_debug ("running attachments job");
 
-	pps_document_doc_mutex_lock ();
+	pps_document_doc_mutex_lock (pps_job_get_document (job));
 	job_attachments->attachments =
-		pps_document_attachments_get_attachments (PPS_DOCUMENT_ATTACHMENTS (job->document));
-	pps_document_doc_mutex_unlock ();
+		pps_document_attachments_get_attachments (PPS_DOCUMENT_ATTACHMENTS (pps_job_get_document (job)));
+	pps_document_doc_mutex_unlock (pps_job_get_document (job));
 
 	pps_job_succeeded (job);
 
@@ -523,20 +279,20 @@ pps_job_annots_run (PpsJob *job)
 
 	g_debug ("running annots job");
 
-	pps_document_doc_mutex_lock ();
-	for (i = 0; i < pps_document_get_n_pages (job->document); i++) {
+	pps_document_doc_mutex_lock (pps_job_get_document (job));
+	for (i = 0; i < pps_document_get_n_pages (pps_job_get_document (job)); i++) {
 		PpsMappingList *mapping_list;
 		PpsPage        *page;
 
-		page = pps_document_get_page (job->document, i);
-		mapping_list = pps_document_annotations_get_annotations (PPS_DOCUMENT_ANNOTATIONS (job->document),
+		page = pps_document_get_page (pps_job_get_document (job), i);
+		mapping_list = pps_document_annotations_get_annotations (PPS_DOCUMENT_ANNOTATIONS (pps_job_get_document (job)),
 									page);
 		g_object_unref (page);
 
 		if (mapping_list)
 			job_annots->annots = g_list_prepend (job_annots->annots, mapping_list);
 	}
-	pps_document_doc_mutex_unlock ();
+	pps_document_doc_mutex_unlock (pps_job_get_document (job));
 
 	job_annots->annots = g_list_reverse (job_annots->annots);
 
@@ -570,6 +326,18 @@ pps_job_annots_new (PpsDocument *document)
 	return job;
 }
 
+/**
+ * pps_job_annots_get_annots:
+ * @job: an #PpsJobAnnots
+ *
+ * Returns: (transfer none) (element-type PpsMappingList):
+ */
+GList *
+pps_job_annots_get_annots (PpsJobAnnots *job)
+{
+	return job->annots;
+}
+
 /* PpsJobRenderTexture */
 static void
 pps_job_render_texture_init (PpsJobRenderTexture *job)
@@ -625,23 +393,21 @@ pps_job_render_texture_run (PpsJob *job)
 
 	g_debug ("running render job: page: %d (%p)", job_render->page, job);
 
-	pps_document_doc_mutex_lock ();
-	pps_document_fc_mutex_lock ();
+	pps_document_doc_mutex_lock (pps_job_get_document (job));
 
 	PPS_PROFILER_START (PPS_GET_TYPE_NAME (job), g_strdup_printf("page: %d", job_render->page));
 
-	pps_page = pps_document_get_page (job->document, job_render->page);
+	pps_page = pps_document_get_page (pps_job_get_document (job), job_render->page);
 	rc = pps_render_context_new (pps_page, job_render->rotation, job_render->scale);
 	pps_render_context_set_target_size (rc,
 					   job_render->target_width, job_render->target_height);
 	g_object_unref (pps_page);
 
-	surface = pps_document_render (job->document, rc);
+	surface = pps_document_render (pps_job_get_document (job), rc);
 
 	if (surface == NULL ||
 	    cairo_surface_status (surface) != CAIRO_STATUS_SUCCESS) {
-		pps_document_fc_mutex_unlock ();
-		pps_document_doc_mutex_unlock ();
+		pps_document_doc_mutex_unlock (pps_job_get_document (job));
 		g_object_unref (rc);
 
                 if (surface != NULL) {
@@ -670,16 +436,15 @@ pps_job_render_texture_run (PpsJob *job)
 	/* If job was cancelled during the page rendering,
 	 * we return now, so that the thread is finished ASAP
 	 */
-	if (g_cancellable_is_cancelled (job->cancellable)) {
+	if (g_cancellable_is_cancelled (pps_job_get_cancellable (job))) {
 		PPS_PROFILER_STOP ();
-		pps_document_fc_mutex_unlock ();
-		pps_document_doc_mutex_unlock ();
+		pps_document_doc_mutex_unlock (pps_job_get_document (job));
 		g_object_unref (rc);
 		return FALSE;
 	}
 
-	if (job_render->include_selection && PPS_IS_SELECTION (job->document)) {
-		pps_selection_render_selection (PPS_SELECTION (job->document),
+	if (job_render->include_selection && PPS_IS_SELECTION (pps_job_get_document (job))) {
+		pps_selection_render_selection (PPS_SELECTION (pps_job_get_document (job)),
 					       rc,
 					       &selection,
 					       &(job_render->selection_points),
@@ -687,7 +452,7 @@ pps_job_render_texture_run (PpsJob *job)
 					       job_render->selection_style,
 					       &(job_render->text), &(job_render->base));
 		job_render->selection_region =
-			pps_selection_get_selection_region (PPS_SELECTION (job->document),
+			pps_selection_get_selection_region (PPS_SELECTION (pps_job_get_document (job)),
 							   rc,
 							   job_render->selection_style,
 							   &(job_render->selection_points));
@@ -701,8 +466,7 @@ pps_job_render_texture_run (PpsJob *job)
 	g_object_unref (rc);
 
 	PPS_PROFILER_STOP ();
-	pps_document_fc_mutex_unlock ();
-	pps_document_doc_mutex_unlock ();
+	pps_document_doc_mutex_unlock (pps_job_get_document (job));
 
 	pps_job_succeeded (job);
 
@@ -748,23 +512,23 @@ pps_job_page_data_run (PpsJob *job)
 
 	g_debug ("running page data job: page: %d (%p)", job_pd->page, job);
 
-	pps_document_doc_mutex_lock ();
-	pps_page = pps_document_get_page (job->document, job_pd->page);
+	pps_document_doc_mutex_lock (pps_job_get_document (job));
+	pps_page = pps_document_get_page (pps_job_get_document (job), job_pd->page);
 
-	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_TEXT_MAPPING) && PPS_IS_DOCUMENT_TEXT (job->document))
+	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_TEXT_MAPPING) && PPS_IS_DOCUMENT_TEXT (pps_job_get_document (job)))
 		job_pd->text_mapping =
-			pps_document_text_get_text_mapping (PPS_DOCUMENT_TEXT (job->document), pps_page);
-	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_TEXT) && PPS_IS_DOCUMENT_TEXT (job->document))
+			pps_document_text_get_text_mapping (PPS_DOCUMENT_TEXT (pps_job_get_document (job)), pps_page);
+	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_TEXT) && PPS_IS_DOCUMENT_TEXT (pps_job_get_document (job)))
 		job_pd->text =
-			pps_document_text_get_text (PPS_DOCUMENT_TEXT (job->document), pps_page);
-	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_TEXT_LAYOUT) && PPS_IS_DOCUMENT_TEXT (job->document))
-		pps_document_text_get_text_layout (PPS_DOCUMENT_TEXT (job->document),
+			pps_document_text_get_text (PPS_DOCUMENT_TEXT (pps_job_get_document (job)), pps_page);
+	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_TEXT_LAYOUT) && PPS_IS_DOCUMENT_TEXT (pps_job_get_document (job)))
+		pps_document_text_get_text_layout (PPS_DOCUMENT_TEXT (pps_job_get_document (job)),
 						  pps_page,
 						  &(job_pd->text_layout),
 						  &(job_pd->text_layout_length));
-	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_TEXT_ATTRS) && PPS_IS_DOCUMENT_TEXT (job->document))
+	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_TEXT_ATTRS) && PPS_IS_DOCUMENT_TEXT (pps_job_get_document (job)))
 		job_pd ->text_attrs =
-			pps_document_text_get_text_attrs (PPS_DOCUMENT_TEXT (job->document),
+			pps_document_text_get_text_attrs (PPS_DOCUMENT_TEXT (pps_job_get_document (job)),
 							 pps_page);
         if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_TEXT_LOG_ATTRS) && job_pd->text) {
                 job_pd->text_log_attrs_length = g_utf8_strlen (job_pd->text, -1);
@@ -773,27 +537,27 @@ pps_job_page_data_run (PpsJob *job)
                 /* FIXME: We need API to get the language of the document */
                 pango_get_log_attrs (job_pd->text, -1, -1, NULL, job_pd->text_log_attrs, job_pd->text_log_attrs_length + 1);
         }
-	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_LINKS) && PPS_IS_DOCUMENT_LINKS (job->document))
+	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_LINKS) && PPS_IS_DOCUMENT_LINKS (pps_job_get_document (job)))
 		job_pd->link_mapping =
-			pps_document_links_get_links (PPS_DOCUMENT_LINKS (job->document), pps_page);
-	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_FORMS) && PPS_IS_DOCUMENT_FORMS (job->document))
+			pps_document_links_get_links (PPS_DOCUMENT_LINKS (pps_job_get_document (job)), pps_page);
+	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_FORMS) && PPS_IS_DOCUMENT_FORMS (pps_job_get_document (job)))
 		job_pd->form_field_mapping =
-			pps_document_forms_get_form_fields (PPS_DOCUMENT_FORMS (job->document),
+			pps_document_forms_get_form_fields (PPS_DOCUMENT_FORMS (pps_job_get_document (job)),
 							   pps_page);
-	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_IMAGES) && PPS_IS_DOCUMENT_IMAGES (job->document))
+	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_IMAGES) && PPS_IS_DOCUMENT_IMAGES (pps_job_get_document (job)))
 		job_pd->image_mapping =
-			pps_document_images_get_image_mapping (PPS_DOCUMENT_IMAGES (job->document),
+			pps_document_images_get_image_mapping (PPS_DOCUMENT_IMAGES (pps_job_get_document (job)),
 							      pps_page);
-	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_ANNOTS) && PPS_IS_DOCUMENT_ANNOTATIONS (job->document))
+	if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_ANNOTS) && PPS_IS_DOCUMENT_ANNOTATIONS (pps_job_get_document (job)))
 		job_pd->annot_mapping =
-			pps_document_annotations_get_annotations (PPS_DOCUMENT_ANNOTATIONS (job->document),
+			pps_document_annotations_get_annotations (PPS_DOCUMENT_ANNOTATIONS (pps_job_get_document (job)),
 								 pps_page);
-        if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_MEDIA) && PPS_IS_DOCUMENT_MEDIA (job->document))
+        if ((job_pd->flags & PPS_PAGE_DATA_INCLUDE_MEDIA) && PPS_IS_DOCUMENT_MEDIA (pps_job_get_document (job)))
                 job_pd->media_mapping =
-                        pps_document_media_get_media_mapping (PPS_DOCUMENT_MEDIA (job->document),
+                        pps_document_media_get_media_mapping (PPS_DOCUMENT_MEDIA (pps_job_get_document (job)),
                                                              pps_page);
 	g_object_unref (pps_page);
-	pps_document_doc_mutex_unlock ();
+	pps_document_doc_mutex_unlock (pps_job_get_document (job));
 
 	pps_job_succeeded (job);
 
@@ -855,22 +619,22 @@ pps_job_thumbnail_texture_run (PpsJob *job)
 
 	g_debug ("running thumbnail job: page: %d (%p)", job_thumb->page, job);
 
-	pps_document_doc_mutex_lock ();
+	pps_document_doc_mutex_lock (pps_job_get_document (job));
 
 	PPS_PROFILER_START (PPS_GET_TYPE_NAME (job), g_strdup_printf("page: %d", job_thumb->page));
-	page = pps_document_get_page (job->document, job_thumb->page);
+	page = pps_document_get_page (pps_job_get_document (job), job_thumb->page);
 	rc = pps_render_context_new (page, job_thumb->rotation, job_thumb->scale);
 	pps_render_context_set_target_size (rc,
 					   job_thumb->target_width, job_thumb->target_height);
 	g_object_unref (page);
 
-	surface = pps_document_get_thumbnail_surface (job->document, rc);
+	surface = pps_document_get_thumbnail_surface (pps_job_get_document (job), rc);
 
 	job_thumb->thumbnail_texture = pps_document_misc_texture_from_surface (surface);
 	cairo_surface_destroy(surface);
 	g_object_unref (rc);
 	PPS_PROFILER_STOP ();
-	pps_document_doc_mutex_unlock ();
+	pps_document_doc_mutex_unlock (pps_job_get_document (job));
 
 	if (job_thumb->thumbnail_texture == NULL) {
 		pps_job_failed (job,
@@ -964,13 +728,11 @@ pps_job_fonts_run (PpsJob *job)
 
 	g_debug ("running fonts job");
 
-	pps_document_doc_mutex_lock ();
-	pps_document_fc_mutex_lock ();
+	pps_document_doc_mutex_lock (pps_job_get_document (job));
 
 	pps_document_fonts_scan (PPS_DOCUMENT_FONTS (document));
 
-	pps_document_fc_mutex_unlock ();
-	pps_document_doc_mutex_unlock ();
+	pps_document_doc_mutex_unlock (pps_job_get_document (job));
 
 	pps_job_succeeded (job);
 
@@ -1069,8 +831,6 @@ pps_job_load_run (PpsJob *job)
 		return FALSE;
 	}
 
-	pps_document_fc_mutex_lock ();
-
 	/* This job may already have a document even if the job didn't complete
 	   because, e.g., a password is required - if so, just reload rather than
 	   creating a new instance */
@@ -1083,9 +843,7 @@ pps_job_load_run (PpsJob *job)
 							   job_load->password);
 		}
 
-		job->failed = FALSE;
-		job->finished = FALSE;
-		g_clear_error (&job->error);
+		pps_job_reset (job);
 
 		if (job_load->uri) {
 			uncompressed_uri = g_object_get_data (G_OBJECT (loaded_doc),
@@ -1104,7 +862,7 @@ pps_job_load_run (PpsJob *job)
 				pps_document_load_fd (loaded_doc,
 						     fd,
 						     job_load->flags,
-						     job->cancellable,
+						     pps_job_get_cancellable (job),
 						     &error);
 		}
 	} else {
@@ -1124,13 +882,11 @@ pps_job_load_run (PpsJob *job)
 					pps_document_factory_get_document_for_fd (fd,
 										 job_load->mime_type,
 										 job_load->flags,
-										 job->cancellable,
+										 pps_job_get_cancellable (job),
 										 &error);
 		}
 	}
 
-	pps_document_fc_mutex_unlock ();
-
 	if (error) {
 		pps_job_failed_from_error (job, error);
 		g_error_free (error);
@@ -1254,6 +1010,12 @@ pps_job_load_take_fd (PpsJobLoad  *job,
 	job->fd = fd;
 }
 
+/**
+ * pps_job_load_set_password:
+ * @job: an #PpsJobLoad
+ * @password: (nullable): the password
+ *
+ */
 void
 pps_job_load_set_password (PpsJobLoad *job, const gchar *password)
 {
@@ -1265,6 +1027,32 @@ pps_job_load_set_password (PpsJobLoad *job, const gchar *password)
 	job->password = g_strdup (password);
 }
 
+const gchar *
+pps_job_load_get_password (PpsJobLoad *job)
+{
+	g_return_val_if_fail (PPS_IS_JOB_LOAD (job), NULL);
+
+	return job->password;
+}
+
+void
+pps_job_load_set_password_save (PpsJobLoad *job, GPasswordSave save)
+{
+	g_return_if_fail (PPS_IS_JOB_LOAD (job));
+
+	g_debug ("load job setting password save");
+
+	job->password_save = save;
+}
+
+GPasswordSave
+pps_job_load_get_password_save (PpsJobLoad *job)
+{
+	g_return_val_if_fail (PPS_IS_JOB_LOAD (job), G_PASSWORD_SAVE_NEVER);
+
+	return job->password_save;
+}
+
 void
 pps_job_load_set_load_flags (PpsJobLoad           *job,
 			    PpsDocumentLoadFlags  flags)
@@ -1294,6 +1082,16 @@ pps_job_load_get_loaded_document (PpsJobLoad *job)
 }
 
 /* PpsJobSave */
+
+typedef struct _PpsJobSavePrivate {
+	gchar *uri;
+	gchar *document_uri;
+} PpsJobSavePrivate;
+
+G_DEFINE_TYPE_WITH_PRIVATE (PpsJobSave, pps_job_save, PPS_TYPE_JOB)
+
+#define JOB_SAVE_GET_PRIVATE(o) pps_job_save_get_instance_private (o)
+
 static void
 pps_job_save_init (PpsJobSave *job)
 {
@@ -1302,27 +1100,29 @@ pps_job_save_init (PpsJobSave *job)
 static void
 pps_job_save_dispose (GObject *object)
 {
-	PpsJobSave *job = PPS_JOB_SAVE (object);
+	PpsJobSave *job_save = PPS_JOB_SAVE (object);
+	PpsJobSavePrivate *priv = JOB_SAVE_GET_PRIVATE(job_save);
 
-	g_debug ("disposing job save: uri: %s", job->uri);
+	g_debug ("disposing job save: uri: %s", priv->uri);
 
-	g_clear_pointer (&job->uri, g_free);
-	g_clear_pointer (&job->document_uri, g_free);
+	g_clear_pointer (&priv->uri, g_free);
+	g_clear_pointer (&priv->document_uri, g_free);
 
-	(* G_OBJECT_CLASS (pps_job_save_parent_class)->dispose) (object);
+	G_OBJECT_CLASS (pps_job_save_parent_class)->dispose (object);
 }
 
 static gboolean
 pps_job_save_run (PpsJob *job)
 {
 	PpsJobSave *job_save = PPS_JOB_SAVE (job);
+	PpsJobSavePrivate *priv = JOB_SAVE_GET_PRIVATE(job_save);
 	gint       fd;
 	gchar     *tmp_filename = NULL;
 	gchar     *local_uri;
 	GError    *error = NULL;
 
 	g_debug ("running save job: uri: %s, document_uri: %s",
-			   job_save->uri, job_save->document_uri);
+			   priv->uri, priv->document_uri);
 
         fd = pps_mkstemp ("saveacopy.XXXXXX", &tmp_filename, &error);
         if (fd == -1) {
@@ -1333,15 +1133,15 @@ pps_job_save_run (PpsJob *job)
 	}
 	close (fd);
 
-	pps_document_doc_mutex_lock ();
+	pps_document_doc_mutex_lock (pps_job_get_document (job));
 
 	/* Save document to temp filename */
 	local_uri = g_filename_to_uri (tmp_filename, NULL, &error);
         if (local_uri != NULL) {
-                pps_document_save (job->document, local_uri, &error);
+                pps_document_save (pps_job_get_document (job), local_uri, &error);
         }
 
-	pps_document_doc_mutex_unlock ();
+	pps_document_doc_mutex_unlock (pps_job_get_document (job));
 
 	if (error) {
 		g_free (local_uri);
@@ -1354,16 +1154,16 @@ pps_job_save_run (PpsJob *job)
 	/* If original document was compressed,
 	 * compress it again before saving
 	 */
-	if (g_object_get_data (G_OBJECT (job->document), "uri-uncompressed")) {
+	if (g_object_get_data (G_OBJECT (pps_job_get_document (job)), "uri-uncompressed")) {
 		PpsCompressionType ctype = PPS_COMPRESSION_NONE;
 		const gchar      *ext;
 		gchar            *uri_comp;
 
-		ext = g_strrstr (job_save->document_uri, ".gz");
+		ext = g_strrstr (priv->document_uri, ".gz");
 		if (ext && g_ascii_strcasecmp (ext, ".gz") == 0)
 			ctype = PPS_COMPRESSION_GZIP;
 
-		ext = g_strrstr (job_save->document_uri, ".bz2");
+		ext = g_strrstr (priv->document_uri, ".bz2");
 		if (ext && g_ascii_strcasecmp (ext, ".bz2") == 0)
 			ctype = PPS_COMPRESSION_BZIP2;
 
@@ -1391,13 +1191,13 @@ pps_job_save_run (PpsJob *job)
 	if (!local_uri)
 		return FALSE;
 
-	pps_xfer_uri_simple (local_uri, job_save->uri, &error);
+	pps_xfer_uri_simple (local_uri, priv->uri, &error);
 	pps_tmp_uri_unlink (local_uri);
 
         /* Copy the metadata from the original file */
         if (!error) {
                 /* Ignore errors here. Failure to copy metadata is not a hard error */
-                pps_file_copy_metadata (job_save->document_uri, job_save->uri, NULL);
+                pps_file_copy_metadata (priv->document_uri, priv->uri, NULL);
         }
 
 	if (error) {
@@ -1426,6 +1226,7 @@ pps_job_save_new (PpsDocument  *document,
 		 const gchar *document_uri)
 {
 	PpsJobSave *job;
+	PpsJobSavePrivate *priv;
 
 	g_debug ("new save job: uri: %s, document_uri: %s", uri, document_uri);
 
@@ -1433,12 +1234,22 @@ pps_job_save_new (PpsDocument  *document,
 			    "document", document,
 			    NULL);
 
-	job->uri = g_strdup (uri);
-	job->document_uri = g_strdup (document_uri);
+	priv = JOB_SAVE_GET_PRIVATE(job);
+
+	priv->uri = g_strdup (uri);
+	priv->document_uri = g_strdup (document_uri);
 
 	return PPS_JOB (job);
 }
 
+const gchar *
+pps_job_save_get_uri (PpsJobSave *job_save)
+{
+	PpsJobSavePrivate *priv = JOB_SAVE_GET_PRIVATE(job_save);
+
+	return priv->uri;
+}
+
 /* PpsJobFind */
 static void
 pps_job_find_init (PpsJobFind *job)
@@ -1469,29 +1280,33 @@ static gboolean
 pps_job_find_run (PpsJob *job)
 {
 	PpsJobFind      *job_find = PPS_JOB_FIND (job);
-	PpsDocumentFind *find = PPS_DOCUMENT_FIND (job->document);
+	PpsDocumentFind *find = PPS_DOCUMENT_FIND (pps_job_get_document (job));
 	PpsPage         *pps_page;
 	GList           *matches;
+	gint             n_pages, current_page;
 
 	g_debug ("running find job");
 
-	for (gint current_page = job_find->start_page;
-	     (current_page + 1) % job_find->n_pages != job_find->start_page;
-	     current_page = (current_page + 1) % job_find->n_pages) {
-		if (g_cancellable_is_cancelled (job->cancellable))
+	n_pages = job_find->n_pages;
+	current_page = job_find->start_page;
+
+	while (n_pages-- > 0) {
+		if (g_cancellable_is_cancelled (pps_job_get_cancellable (job)))
 			return FALSE;
 
-		pps_document_doc_mutex_lock ();
-		pps_page = pps_document_get_page (job->document, current_page);
+		pps_document_doc_mutex_lock (pps_job_get_document (job));
+		pps_page = pps_document_get_page (pps_job_get_document (job), current_page);
 		matches = pps_document_find_find_text (find, pps_page, job_find->text,
 						       job_find->options);
 		g_object_unref (pps_page);
-		pps_document_doc_mutex_unlock ();
+		pps_document_doc_mutex_unlock (pps_job_get_document (job));
 
 		job_find->has_results |= (matches != NULL);
 
 		job_find->pages[current_page] = matches;
 		g_signal_emit (job_find, job_find_signals[FIND_UPDATED], 0, current_page);
+
+		current_page = (current_page + 1) % job_find->n_pages;
 	}
 
 	pps_job_succeeded (job);
@@ -1624,9 +1439,9 @@ pps_job_layers_run (PpsJob *job)
 
 	g_debug ("running layers job");
 
-	pps_document_doc_mutex_lock ();
-	job_layers->model = pps_document_layers_get_layers (PPS_DOCUMENT_LAYERS (job->document));
-	pps_document_doc_mutex_unlock ();
+	pps_document_doc_mutex_lock (pps_job_get_document (job));
+	job_layers->model = pps_document_layers_get_layers (PPS_DOCUMENT_LAYERS (pps_job_get_document (job)));
+	pps_document_doc_mutex_unlock (pps_job_get_document (job));
 
 	pps_job_succeeded (job);
 
@@ -1697,13 +1512,11 @@ pps_job_export_run (PpsJob *job)
 
 	g_debug ("running export job");
 
-	pps_document_doc_mutex_lock ();
+	pps_document_doc_mutex_lock (pps_job_get_document (job));
 
-	pps_page = pps_document_get_page (job->document, job_export->page);
+	pps_page = pps_document_get_page (pps_job_get_document (job), job_export->page);
 	if (job_export->rc) {
-		job->failed = FALSE;
-		job->finished = FALSE;
-		g_clear_error (&job->error);
+		pps_job_reset (job);
 
 		pps_render_context_set_page (job_export->rc, pps_page);
 	} else {
@@ -1711,9 +1524,9 @@ pps_job_export_run (PpsJob *job)
 	}
 	g_object_unref (pps_page);
 
-	pps_file_exporter_do_page (PPS_FILE_EXPORTER (job->document), job_export->rc);
+	pps_file_exporter_do_page (PPS_FILE_EXPORTER (pps_job_get_document (job)), job_export->rc);
 
-	pps_document_doc_mutex_unlock ();
+	pps_document_doc_mutex_unlock (pps_job_get_document (job));
 
 	pps_job_succeeded (job);
 
@@ -1781,20 +1594,18 @@ pps_job_print_run (PpsJob *job)
 
 	g_debug ("running print job");
 
-	job->failed = FALSE;
-	job->finished = FALSE;
-	g_clear_error (&job->error);
+	pps_job_reset (job);
 
-	pps_document_doc_mutex_lock ();
+	pps_document_doc_mutex_lock (pps_job_get_document (job));
 
-	pps_page = pps_document_get_page (job->document, job_print->page);
-	pps_document_print_print_page (PPS_DOCUMENT_PRINT (job->document),
+	pps_page = pps_document_get_page (pps_job_get_document (job), job_print->page);
+	pps_document_print_print_page (PPS_DOCUMENT_PRINT (pps_job_get_document (job)),
 				      pps_page, job_print->cr);
 	g_object_unref (pps_page);
 
-	pps_document_doc_mutex_unlock ();
+	pps_document_doc_mutex_unlock (pps_job_get_document (job));
 
-        if (g_cancellable_is_cancelled (job->cancellable))
+        if (g_cancellable_is_cancelled (pps_job_get_cancellable (job)))
                 return FALSE;
 
 	cr_status = cairo_status (job_print->cr);
