@@ -25,7 +25,6 @@
 #endif
 
 #include "pps-view.h"
-#include "pps-document-model.h"
 #include "pps-pixbuf-cache.h"
 #include "pps-page-cache.h"
 #include "pps-jobs.h"
@@ -33,8 +32,7 @@
 #include "pps-form-field.h"
 #include "pps-selection.h"
 #include "pps-view-cursor.h"
-
-#define DRAG_HISTORY 10
+#include "adwaita.h"
 
 struct GdkPoint {
 	gint x;
@@ -45,38 +43,20 @@ typedef struct GdkPoint GdkPoint;
 
 /* Information for middle clicking and moving around the doc */
 typedef struct {
-        gboolean in_drag;
-	GdkPoint start;
 	gdouble hadj;
 	gdouble vadj;
-	guint drag_timeout_id;
 	guint release_timeout_id;
-	GdkPoint buffer[DRAG_HISTORY];
 	GdkPoint momentum;
 	gboolean in_notify;
 } DragInfo;
 
 /* Information for handling selection */
 typedef struct {
-	gboolean in_select;
 	GdkPoint start;
 	GList *selections;
 	PpsSelectionStyle style;
 } SelectionInfo;
 
-/* Information for handling images DND */
-typedef struct {
-	gboolean in_drag;
-	GdkPoint start;
-	PpsImage *image;
-} ImageDNDInfo;
-
-typedef enum {
-	PPS_PAN_ACTION_NONE,
-	PPS_PAN_ACTION_NEXT,
-	PPS_PAN_ACTION_PREV
-} PpsPanAction;
-
 /* Annotation popup windows */
 typedef struct {
 	GtkWidget *window;
@@ -101,10 +81,7 @@ typedef struct _PpsHeightToPageCache {
 
 /* Information for handling annotations */
 typedef struct {
-	GdkPoint      start;
 	PpsPoint       cursor_offset;
-	gboolean      annot_clicked;
-	gboolean      moving_annot;
 	PpsAnnotation *annot;
 } MovingAnnotInfo;
 
@@ -118,10 +95,18 @@ typedef struct {
 	guint      delay_timeout_id;
 } PpsLinkPreview;
 
+typedef struct {
+	gboolean active;
+	gboolean in_selection;
+	GdkPoint start;
+	GdkPoint stop;
+} SigningInfo;
+
 typedef struct _PpsViewPrivate {
 	PpsDocument *document;
 
 	/* Find */
+	PpsSearchContext *search_context;
 	PpsJobFind *find_job;
 	GList **find_pages; /* Backwards compatibility. Contains PpsFindRectangles's elements per page */
 	gint find_page;     /* Page of active find result */
@@ -158,6 +143,11 @@ typedef struct _PpsViewPrivate {
 	gboolean      pending_resize;
 	PpsPoint       pending_point;
 
+	/* Animation for scrolling with keys */
+	AdwAnimation	*scroll_animation_vertical;
+	AdwAnimation	*scroll_animation_horizontal;
+	gboolean	pending_scroll_animation;
+
 	/* Current geometry */
 
 	gint start_page;
@@ -179,9 +169,6 @@ typedef struct _PpsViewPrivate {
 	guint loading_timeout;
 	gboolean allow_links_change_zoom;
 
-	/* Common for button press handling */
-	int pressed_button;
-
 	/* Key bindings propagation */
 	gboolean key_binding_handled;
 
@@ -199,7 +186,7 @@ typedef struct _PpsViewPrivate {
 	PpsLinkAction *link_selected;
 
 	/* Image DND */
-	ImageDNDInfo image_dnd_info;
+	PpsImage *dnd_image;
 
 	/* Annotations */
 	GList             *window_children;
@@ -223,10 +210,9 @@ typedef struct _PpsViewPrivate {
 	guint    cursor_blink_time;
 
 	/* Gestures */
-	GtkGesture *pan_gesture;
-	GtkGesture *zoom_gesture;
+	GtkGesture *middle_clicked_drag_gesture;
+	GtkGesture *middle_clicked_drag_swipe_gesture;
 	gdouble prev_zoom_gesture_scale;
-	PpsPanAction pan_action;
 
 	/* Current zoom center */
 	gdouble zoom_center_x;
@@ -234,6 +220,9 @@ typedef struct _PpsViewPrivate {
 
 	/* Link preview */
 	PpsLinkPreview link_preview;
+
+	/* Signing Info */
+	SigningInfo signing_info;
 } PpsViewPrivate;
 
 struct _PpsViewClass {
@@ -263,6 +252,9 @@ struct _PpsViewClass {
 				       gint            count,
 				       gboolean        extend_selection);
 	void     (*activate)          (PpsView         *view);
+	void     (*signature_rect)    (PpsView         *view,
+				       guint            page,
+				       PpsRectangle    *rectangle);
 };
 
 void _get_page_size_for_scale_and_rotation (PpsDocument *document,
