@@ -1,16 +1,13 @@
 #![cfg_attr(docsrs, feature(doc_cfg))]
 
-/// No-op.
-macro_rules! skip_assert_initialized {
-    () => {};
-}
-
 /// No-op.
 macro_rules! assert_initialized_main_thread {
     () => {};
 }
 
+pub use auto::functions::*;
 pub use auto::*;
+pub use ffi;
 #[allow(unused_imports)]
 mod auto;
 
